#include "sdb.h"

bool rank_segment(segment s1, segment s2)
{
	if (Tools::double_eq(s1.start.x, s2.start.x, 6))
		return (Tools::double_le(s2.end.x, s1.end.x, 6));
	return Tools::double_le(s2.start.x, s1.start.x,6);
}


SDB::SDB(char* datafile)
{
	N=0;
	M=0;
	this->datafile=datafile;
	fin.open(datafile);
	if (!fin)
		cerr<<"cannot open data file "<<datafile<<endl;

	ymax=-1.0;
	ymin=100000000;
 	xmin=10000000;
	xmax=-1;
}


SDB::~SDB()
{
	if (fin.is_open())
		fin.close();
	segs.clear();
}

void SDB::readEasyData()
{
	if (!fin)
		return;
	double oid, sid;
	fin>>M>>N;
	cout<<M<<" "<<N<<endl;
	segment t;
	point p1,p2;
	for (uint i=0;i<N;i++)
	{
		fin>>oid>>sid>>p1>>p2;
		t.start=p1;
		t.end=p2;
		t.objid=(int)oid;
		t.segid=(int)sid;
		t.buildseg();	
		segs.push_back(t);
	}
}

void SDB::readData()
{
	if (!fin)
		return;
	fin>>M;
	//cout<<"M is "<<M<<endl;
	uint n,i,j;
	uint oid=0, sid=0;
	double x=0.0, y=0.0;
	j=0;
	vector<point> P;
	while(j<M)
	{
		if (!fin.good()) continue;
		fin>>n; 
		//	cout<<"n is "<<n<<endl;
		P.clear();
		P.resize(n);
		sid=0;
		for (i=0;i<n;i++)
		{
			fin>>x>>y;
			P[i].x=x;
			P[i].y=y;
			if (xmin>x)
				xmin=x;
			if (xmax<x)
				xmax=x;
			if (ymin>y)
				ymin=y;
			if (ymax<y)
				ymax=y;
		}
		for (i=1;i<n;i++)
		{
			segment s;
			s.buildseg(P[i-1],P[i], oid, sid);
			sid++; 
			segs.push_back(s);		
		}
		j++;
		oid++;
	}
	N=segs.size();
} 


void SDB::sortSegments()
{
	sort(segs.begin(),segs.end(),rank_segment);
}

void SDB::sample(double rate, vector<segment>& spdSegs, vector<uint>& uspdSegs)
{
	uint i=0;
	double val=0.0;
	for(i=0;i<N;i++)
	{
		val=rng.uniform(0.0,1.0);
		if (val<rate)
			spdSegs.push_back(segs[i]);
		else
			uspdSegs.push_back(i);
	}	
	sort(spdSegs.begin(),spdSegs.end(),rank_segment);
}

void SDB::printSegments()
{
	uint i=0;
	cout<<M<<" "<<N<<" "<<xmin<<" "<<xmax<<" "<<ymin<<" "<<ymax<<endl;
	for (i=0;i<N;i++)
		cout<<segs[i]<<endl;
}

void SDB::printSegments(vector<segment>& segs)
{
	for (uint i=0;i<segs.size();i++)
		cout<<segs[i]<<endl;
}

void SDB::printSegments(list<segment>& segs)
{
	list<segment>::iterator it;
	for (it=segs.begin();it!=segs.end();it++)
		cout<<(*it)<<endl;
}

void SDB::printSegments(vector<segment>& segs, vector<uint> segIndex)
{
	for (uint i=0;i<segIndex.size();i++)
		cout<<segs[segIndex[i]]<<endl;
}

void SDB::printSegments(vector<segment>& segs, char* file)
{

	fstream fs;
	fs.open(file, ios::out);

	for (uint i=0;i<segs.size();i++)
		fs<<segs[i]<<endl;
	fs.close();	
}

void SDB::printSegments(list<segment>& segs, char* file)
{
	fstream fs;
	fs.open(file,ios::out);
	list<segment>::iterator it;

	for (it=segs.begin();it!=segs.end();it++)
		fs<<(*it)<<endl;
	fs.close();
}

void SDB::printSegments(vector<segment>& segs, vector<uint> segIndex, char* file)
{
	fstream fs;
	fs.open(file,ios::out);
	for (uint i=0;i<segIndex.size();i++)
		fs<<segs[segIndex[i]]<<endl;
	fs.close();
}
uint SDB::outputConflictList(vector<segment>& segs, vector<uint>& segIndex, char* ofile)
{
  fstream ofs;
  ofs.open(ofile, ios::binary|ios::out|ios::app);
  uint ppos=ofs.tellg();
  uint index=0;
  uint ssize=segIndex.size();
  //shortsegment st;
  for(uint i=0;i<ssize;i++)
    {
      index=segIndex[i];
      //st.buildseg(segs[index].start, segs[index].end, segs[index].objid, segs[index].segid);
      //ofs.write((char*)( &(st) ),sizeof(shortsegment));
      ofs.write((char*)( &(segs[index]) ),(sizeof(segment)-2*sizeof(double)) );
    }
  ofs.close();
  return ppos;
}

uint SDB::readConflictList(vector<segment>& segs, char* ifile, uint ppos, uint csize)
{
	fstream r;
	r.open(ifile,ios::binary|ios::in);
	r.seekg(ppos);
	segment t;
	uint count=0;
	//shortsegment st;
	while(csize)
  	{
	    r.read((char*)&t,(sizeof(segment)-2*sizeof(double)));
	    t.buildseg();
	    segs.push_back(t);
	    count++;
	    csize--;
        }
	r.close();
	//segs.pop_back(); //hadle the c++ stupid bug: reading the last line twice in a stream 
	//count--;
	return count;
}

/*
void SDB::readNConflictList(vector<segment>& segs, vector<uint>& clevel, vector<uint>&  cindex, char* filepath, uint& cios, uint B)
{
	uint i=0;
	char buf[100]="";
	uint count=0;
	for (i=0;i<clevel.size();i++)
	{
		sprintf(buf, "%sS%d/C/C%d",filepath, clevel[i],cindex[i]);
		count+=SDB::readConflictList(segs, buf);
		//cout<<buf<<" "<<count<<endl;
	}
	cios+=(uint)ceil((double)count/(double)B);
}
*/

void SDB::readEnvelop(const char* envfile, vector <segment> & s,vector <unsigned long int> & cn, vector<uint>& cd, vector<uint>& ppos)//low:high|x:small->large
  {
    segment t;
    unsigned long int n=0;
	uint distinct=0;
	uint pos=0;
    ifstream is;
    is.open(envfile);
    while(!is.eof())
      {
	is>>t>>n>>distinct>>pos;
	s.push_back(t);
	cn.push_back(n);
	cd.push_back(distinct);
	ppos.push_back(pos);
      }
     is.close();
     s.pop_back();
     cn.pop_back(); //hadle the c++ stupid bug: reading the last line twice in a stream 
     cd.pop_back();
  }

void SDB::readREnvelop(const char* envfile, vector <segment> & s,vector <uint> & cn,  vector <uint> & cd, vector<uint>& depend, vector <uint> & clevel, vector<uint>& cindex)
{
  segment t;
  uint n=0,distinct=0, dependnum=0, index=0, level=0;
  ifstream is;
  uint total=0;
  is.open(envfile);
  is>>total; 
  for (uint i=0;i<total;i++)
    {
      is>>t>>n>>distinct>>dependnum;
      s.push_back(t);
      cn.push_back(n);
      cd.push_back(distinct);
      depend.push_back(dependnum);
	for (uint i=0;i<dependnum;i++)
	{
		  is>>level>>index;
		clevel.push_back(level);	
		cindex.push_back(index);
	}
    }
  is.close();
  //s.pop_back();
  //cn.pop_back(); //handle the c++ stupid bug: reading the last line twice in a stream 
/*	if (dependnum!=0)
	{
		clevel.pop_back();
		cindex.pop_back();
	}*/
}


void SDB::printEnvelop(const char* envfile, vector <segment> & s,vector <unsigned long int> & cn)
{
	fstream ofs;
	ofs.open(envfile, ios::out);
	for(uint i=0;i<s.size();i++)
		ofs<<s[i]<<" "<<cn[i]<<endl;
	ofs.close();
}
/*list<segment> SDB::buildEnvelop(vector<segment>& input)
{
	sky s;
	return s.getenvlp(input);
}*/

