//! \file examples/Example_2/ex_envelope_segments.cpp
// Constructing the lower envelope of a set of segments.

#include <CGAL/Gmpq.h>
#include <CGAL/Cartesian.h>
#include <CGAL/Arr_segment_traits_2.h>
#include <CGAL/Arr_curve_data_traits_2.h>
#include <CGAL/Env_default_diagram_1.h>
#include <CGAL/envelope_2.h>

#include <list>
#include <iostream>
#include "sdb.h"


typedef CGAL::Gmpq                                      Number_type;
//typedef double                                      Number_type;
typedef CGAL::Cartesian<Number_type>                    Kernel;
typedef CGAL::Arr_segment_traits_2<Kernel>              Segment_traits_2;
typedef Segment_traits_2::X_monotone_curve_2            Segment_2;
typedef CGAL::Arr_curve_data_traits_2<Segment_traits_2,
                                      uint>             Traits_2;
typedef Traits_2::Point_2                               Point_2;
typedef Traits_2::X_monotone_curve_2                    Labeled_segment_2;
typedef CGAL::Env_default_diagram_1<Traits_2>           Diagram_1;

int main (int argc, char* argv[])
{
  // Consrtuct the input segments and label them 'A' ... 'H'.
  if(argc!=4)
    {
      cout<<"Usage: cglenv data_file rawdata(0:false 1:yes) samlpe_base"<<endl;
      return 0;
    }
	vector<segment> sampled;
	vector<uint> usampled;
	ifstream fin;
	int raw=atoi(argv[2]);
	int base=atoi(argv[3]);
  	std::list<Labeled_segment_2>   segments;

	if (raw==1)
	{	SDB sdb(argv[1]);
		sdb.readData();
		double srate=1.0/(double)base;
		sdb.sample(srate,sampled, usampled); 
	}
	else
	{
		fin.open(argv[1]);
		uint N=0;
		fin>>N;
		sampled.resize(N);
		for (uint i=0;i<N;i++)
		{
			fin>>sampled[i];
			sampled[i].buildseg();
		}
	}
	if (raw==1)
	{
		cerr<<sampled.size()<<endl;
		for (uint i=0;i<sampled.size();i++)
			cerr<<sampled[i]<<endl;
	}		
/*	for (uint i=0;i<2;i++)
 	{ 
		 
	  segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (sampled[i].start.x, sampled[i].start.y),
                                                    Point_2 ( sampled[i].end.x, sampled[i].end.y)), 'B'));
	}*/
	Number_type q(0.5);
	q=0.8;
segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (0, 1),
                                                    Point_2 (q, 3)), 7));
  segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (1, 2),
                                                    Point_2 (4, 5)), 8));
  segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (1, 5),
                                                    Point_2 (7, 2)), 9));
  segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (4, 2),
                                                    Point_2 (6, 4)), 10));
  segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (8, 3),
                                                    Point_2 (8, 6)), 11));
  segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (9, 2),
                                                    Point_2 (12, 4)), 12));
  segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (10, 2),
                                                    Point_2 (12, 1)), 13));
  segments.push_back (Labeled_segment_2 (Segment_2 (Point_2 (11, 0),
                                                    Point_2 (11, 5)), 14));

  // Compute the minimization diagram that represents their lower envelope.
  Diagram_1              min_diag;

  upper_envelope_x_monotone_2 (segments.begin(), segments.end(),  min_diag);
// upper_envelope_2 (segments.begin(), segments.end(),  min_diag);

  // Print the minimization diagram.
  Diagram_1::Edge_const_handle     e = min_diag.leftmost();
  Diagram_1::Vertex_const_handle   v;
  Diagram_1::Curve_const_iterator  cit;
	cout.precision(12);

  while (e != min_diag.rightmost())
  {
    std::cout << "Edge:";
    if (! e->is_empty())
    {
      for (cit = e->curves_begin(); cit != e->curves_end(); ++cit)
      {  	
		std::cout << ' ' << cit->data();
        	uint edgeint= cit->data();
		cout<<"as int "<<edgeint;
	}
    }
    else
      std::cout << " [empty]";
    std::cout << std::endl;

    v = e->right();
    //std::cout << "Vertex (" << v->point() << "):";
	std::cout << "Vertex (" << v->point().x().numerator()<<" "<<v->point().x().denominator()<<" "<< v->point().y().numerator()<<" "<<v->point().y().denominator() << "):";
	double x=CGAL::to_double(v->point().x()); 
	double y=CGAL::to_double(v->point().y()); 
	std::cout<<" "<<x<<" "<<y<<":";	

    for (cit = v->curves_begin(); cit != v->curves_end(); ++cit)
      std::cout << ' ' << cit->data();
    std::cout << std::endl;

    e = v->right();
  }
  CGAL_assertion (e->is_empty());
  std::cout << "Edge: [empty]" << std::endl;
/* while (e != min_diag.rightmost())
  {
   
    v = e->right();
    std::cout <<  v->point();
    for (cit = v->curves_begin(); cit != v->curves_end(); ++cit)
      //std::cout << ' ' << cit->data();
    std::cout << std::endl;

    e = v->right();
  }*/
  CGAL_assertion (e->is_empty());
  std::cout << "Edge: [empty]" << std::endl;
  return (0);
}
