//
// File:    test_ami_btree.cpp
// Author:  Octavian Procopiuc <tavi@cs.duke.edu>
// 
// Test file for class AMI_btree.
//

#include "app_config.h"
#include <portability.h>
#include <versions.h>

/***************************/
#include <sstream>
#include <fstream>
#include <fcntl.h>
#include <unistd.h>
#include <vector>
#include <iostream>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <dirent.h>
#include "mtimer.h"
#include "tools.h"
#include "segment.h"
#include "sdb.h"
/**************************/

VERSION(test_ami_btree_cpp, "$Id: test_ami_btree.cpp,v 1.16 2009/03/19 17:36:13 lifeifei Exp $");

#include <cpu_timer.h>
#include <ami_btree.h>
#include <ami_stream.h>
#define SIZE_OF_STRUCTURE 256

// Key type.
//typedef TPIE_OS_OFFSET bkey_t;
typedef double bkey_t;// le

// Element type for the btree.
struct el_t {
  bkey_t key_;
  uint csize;
  uint cd;
  uint ppos;
  char data_[SIZE_OF_STRUCTURE-sizeof(bkey_t)-3*sizeof(uint)];
  
  el_t(bkey_t k, string d, uint n, uint distinct, uint pos): key_(k) 
  {
    csize=n;
    cd=distinct;
    strcpy(data_,d.data());
    ppos=pos;
  }

  el_t() {}
};

struct key_from_el {
  bkey_t operator()(const el_t& v) const { return v.key_; }
};
// Temporary distinction btw UN*X and WIN, since there are some
// problems with the MMAP collection implementation.
#ifdef _WIN32
typedef AMI_btree< bkey_t,el_t,less<bkey_t>,key_from_el,BTE_COLLECTION_UFS > u_btree_t;
#else
typedef AMI_btree< bkey_t,el_t,less<bkey_t>,key_from_el > u_btree_t;
#endif

typedef AMI_STREAM< el_t > stream_t;

// Template instantiations (to get meaningful output from gprof)
//template class AMI_btree_node<bkey_t,el_t,less<bkey_t>,key_from_el>;
//template class AMI_btree_leaf<bkey_t,el_t,less<bkey_t>,key_from_el>;
//template class AMI_btree< bkey_t,el_t,less<bkey_t>,key_from_el,BTE_COLLECTION_UFS >;
//template class AMI_STREAM< el_t >;


// This is 2**31-1, the max value returned by random().
#define MAX_RANDOM ((double)0x7fffffff)
// This is the max value that we want.
#define MAX_VALUE 1000000000
// The number of deletions to perform.
#define DELETE_COUNT 500

// Global variables.
TPIE_OS_OFFSET bulk_load_count;
TPIE_OS_OFFSET insert_count = 5000;
TPIE_OS_OFFSET range_search_lo = 0;
TPIE_OS_OFFSET range_search_hi = 10000000;

bool rankA(segment s1, segment s2)
{
    if(s1.A<s2.A)
      return true;
    return false;
}


int get_tk(double q, vector <segment> &idlst, int topk, int opflag)
{
  int count=0;
  segment s;
  double x=0,y=0;
  vector <segment> rp;
  for(uint i=0;i<idlst.size();i++)
    {
      s=idlst[i];
      y=0;
      if(Tools::double_le(q, s.start.x, 6)&&Tools::double_le(s.end.x, q,6) && Tools::lineSegmentIntersection(s.start.x,s.start.y,
							     s.end.x,s.end.y,
							     q,2*Tools::MAX,
							     q,0.5,
							     &x,&y)==true)
	{
	  s.A=y;
	  rp.push_back(s);
	  count++; 
	}
    }
  //cerr<<"ok"<<endl<<flush;
  if(count)
    {
	if(rp.size()>0)
	      sort(rp.begin(),rp.end(),rankA);
//	return count;
		
      //report top_k
	if (opflag==1)
	{
	      int i=0;
	      cerr.precision(12);
	      for(i=0;i<rp.size();i++)
		{
	  		  cerr<<rp[i].objid*Tools::SEGIDLEN+rp[i].segid<<" @"<<rp[i]<<" @"<<rp[i].A<<endl;
		  if(i==topk-1)
		    break;
		}
    	}
    }
  rp.clear();
  return count;
}


int get_filesize(const char * filename)
{
  char n[100]="";
  int size=0;
  struct stat buf;
  sprintf(n,"%s",filename);
  if (stat(n, &buf) < 0) {
    fprintf(stderr, "Cannot stat %s\n", n);                                                                               
    exit(1);
  }
  size=buf.st_size;
  return size;
}


int get_treesize(const char * filename)
{
  char n[100]="";
  int size=0;
  struct dirent *w;
  struct stat buf;
  sprintf(n,"%s/btree.ent.l.blk",filename);
  if (stat(n, &buf) < 0) {
    fprintf(stderr, "Cannot stat %s\n", n);                                                                               
    exit(1);
  }
  size+=buf.st_size;
  sprintf(n,"%s/btree.ent.n.blk",filename);
  if (stat(n, &buf) < 0) {
    fprintf(stderr, "Cannot stat %s\n", n);                                                                               
    exit(1);
  }
  size+=buf.st_size;
  return size;
}

int get_cflsize(const char * filename)
{
  char dir[100];
  sprintf(dir,"%s/C",filename);
  DIR *d;
  long size=0;
  struct dirent *w;
  struct stat buf;
  char n[100]="";
  if (stat(dir, &buf) < 0) {
    return 0;
  }
  else
    {
      if(!S_ISDIR(buf.st_mode))
	{
	  if(S_ISLNK(buf.st_mode))
	    {
	      printf("%-ld\t%-s\n",0,dir);
	      exit(0);
	    }
	  else
	    {
	      printf("%-ld\t%-s\n",(long)(buf.st_blocks/2),dir);
	      exit(0);
	    }
	}//if
    }
  if ((d = opendir(dir)) == NULL) {
    printf("This is not supposed to happen.\n");
    exit(0);
  }
  
  while ((w = readdir(d)) != NULL) {
    sprintf(n, "%s/%s", dir, w->d_name);
      if ((strcmp(w->d_name, ".") == 0) ||
	  (strcmp(w->d_name, "..")==0)) continue;
      if (stat(n, &buf) < 0) {
	continue;
      }
      if ((strstr(w->d_name, "."))!=NULL) 
	  continue;
      size+=(long)buf.st_size;
      if (S_ISDIR(buf.st_mode)) {
	size+=get_cflsize(n);
      }
  }
    //stat(dir, &buf);
      //printf("%-ld\t%-s\n",size+(long)(buf.st_blocks/2),dir);
  closedir(d);
  return size;
}


int main(int argc, char *argv[]) {

  AMI_err ae;
  TPIE_OS_OFFSET i;
  TPIE_OS_OFFSET j;
  el_t ss;
  cpu_timer wt;
  int treesize=0,cflsize=0;
  char *base_file = NULL;
  char * base_file_stream = NULL;
  uint T,N=0,k,slvl;
  bool trick_on=false;
  if(argc!=8)
    {
      cerr<<"Usage:query.exe <query input filename> <trick #> <Page Size> <force lvl:S0 is 1> <datafile> <k> <output>"<<endl<<flush;
      exit(0);
    }

  //N=atoi(argv[3]);
  T=atoi(argv[2]);
  /*********build conflict list**************************/

  char datafile_folder[100]="";
  strcpy(datafile_folder,argv[5]);
  strcat(datafile_folder,"Index/");
  char buf[100]="";
  stringstream conv,conv1;
  vector<u_btree_t *> ubt;
  char folder[100]="";
  bool flag=false;
  int pagesize=atoi(argv[3]);
  int Bsize=(int)((double)(pagesize*1024.0)/(double)Tools::OBJSIZE);
   int opflag=atoi(argv[7]);
  //cerr<<Bsize<<endl;
  int force=atoi(argv[4]);
  fstream ifs;
  sprintf(buf,"%sconfig",datafile_folder);
  ifs.open(buf,ios::in);
  if(!ifs.is_open())
    {
      cerr<<"cannot open config file"<<endl;
      exit(1);
    }
  else{
    ifs>>slvl>>N;
    //cerr<<slvl<<" "<<N<<endl;
  }
  ifs.close();
 
  int top_k=atoi(argv[6]);
  long int *hit= new long int[slvl+1];
  memset(hit,0,sizeof(long int)*(slvl+1));
  //cerr<<slvl<<" B+ trees..."<<endl;
  tpie_log_init(TPIE_LOG_APP_DEBUG); 
  //MM_manager.set_memory_limit(64*1024*1024*1024*1024);
  MM_manager.ignore_memory_limit();
  for(k=0;k<slvl;k++)
    {
      conv.str("");
      conv<<datafile_folder;
      conv<<"S"<<k;      
      strcpy(folder,conv.str().data());
      conv.str("");
      conv<<folder<<"/btree.ent"; 
      strcpy(buf,conv.str().data());
      //cerr<<buf<<endl;
      base_file=const_cast <char*> (buf);            
      u_btree_t *u_btree;
      AMI_btree_params params;
      params.node_block_factor = pagesize/4;
      params.leaf_block_factor = pagesize/4;
      params.leaf_cache_size = 32;
      params.node_cache_size = 64;
      u_btree = new u_btree_t(base_file, AMI_WRITE_COLLECTION, params);
      if (!u_btree->is_valid()) {
	cerr << argv[0] << ": Error initializing btree. Aborting.\n";
	delete u_btree;
	exit(1);
      }
      ubt.push_back(u_btree);
    }

  ifstream fs;
  bkey_t key;				 
  unsigned long filesize=0;
  int io_from_tree=0;
  fs.open(argv[1]);
  int count=0, total=0;
  MTimer t;
  uint scanio=(uint)( ceil( (double)N / ((double)Bsize*2.0) ) );
  vector<uint> btreeio;
  vector<uint> cfltio;
  vector<uint> totalio;
  uint btreeioround=0;
  uint cfltioround=0;
  t.go();
  fs>>total;
  while(count<total)
    {
      int num=0;
	btreeioround=0;
	cfltioround=0;
      fs>>key;
      double jdg=((double)top_k/Bsize)>1?((double)top_k/Bsize):1;
      int m=(int)ceil(log(jdg)/log(2));
      segment t;
      vector <segment> vt;
      if (m>force)
	m=force;
      while(num<top_k){
	u_btree_t *u_btree;
	if (m>=ubt.size())
	  {
	    filesize+=scanio;
	    cfltioround+=scanio;
	    hit[slvl]++;
	    break;
	  }
	u_btree=ubt[m];
	if(u_btree->pred(key,ss))				 
	  {
		//cout<<"key "<<key<<"ss.key "<<ss.key_<<" csize "<<ss.csize<<" cd "<<ss.cd<<endl;
	    if(ss.csize<top_k || ( ss.csize>T*Bsize*pow(2.0,m) && T>0 ) )
	      {
		num=-1;		  
		filesize+=u_btree->height();
		io_from_tree+=u_btree->height();
		btreeioround+=u_btree->height();
		if (m==force)
		  {
		    filesize+=scanio;
		     cfltioround+=scanio;
		    hit[m]++;
		    break;
		  }
	      }
	    else
	      {
		vt.clear();
		SDB::readConflictList(vt, ss.data_, ss.ppos, ss.csize);		
		num=vt.size();
		io_from_tree+=u_btree->height();
		filesize+=(uint)ceil((double)(ss.csize)/(double)Bsize)+u_btree->height();

		btreeioround+=u_btree->height();
		cfltioround+=(uint)ceil((double)(ss.csize)/(double)Bsize);

		int a=0;
		a=get_tk(key,vt,top_k,opflag);
		if(a<top_k)
		  {
		    num=-1;
		    if (m==force)
		      {
			filesize+=scanio;
	     	        cfltioround+=scanio;
			hit[m]++;
			break;
		      }
		  }
		else
		  {
		       hit[m]++;
		       break;
		  }
	      }
	  }
	else
	      break;
	m++;
      }
	btreeio.push_back(btreeioround);
	cfltio.push_back(cfltioround);
	totalio.push_back((btreeioround+cfltioround));
      count++;
    }          
  fs.close();
  t.stop();
  t.update();
  //sort(btreeio.begin(), btreeio.end());
  //sort(cfltio.begin(), cfltio.end());
  /*avg time on query, avg cost on tree traverse, avg cost on reading CFL, SUM of the previous two, hit[...], trick, force*/
  cout<<t.elapsed/(double)count<<" "<<(double)io_from_tree/(double)count<<" "<<(double)filesize/(double)count-(double)io_from_tree/(double)count
      <<" "<<(double)filesize/(double)count<<" ";
  int iter=0;
  for(;iter<=ubt.size();iter++)
    cout<<(double)hit[iter]/(double)count<<" ";
  if (ubt.size()<10)
  {
	for (iter=ubt.size()+1;iter<11;iter++)
		cout<<0<<" ";
  }
  cout<<T<<" "<<pagesize<<" "<<force<<" "<<top_k;
  /*ofstream f1, f2;
	if (btreeio.size()>0)
	{
		sort(btreeio.begin(),btreeio.end());
	       sprintf(buf,"%sTreeIOk%dT%d",argv[5],top_k,T);
		f1.open(buf);
		for (uint i=0;i<btreeio.size();i++)
			f1<<btreeio[i]<<endl;	
		f1.close();
	}
	if (cfltio.size()>0)
	{
		sort(cfltio.begin(),cfltio.end());
	       sprintf(buf,"%sCIOk%dT%d",argv[5],top_k,T);
		f2.open(buf);  
		for (uint i=0;i<cfltio.size();i++)
			f2<<cfltio[i]<<endl;	
		f2.close();
	}
	*/
	sort(btreeio.begin(),btreeio.end());
	sort(cfltio.begin(),cfltio.end());
	sort(totalio.begin(),totalio.end());
	uint size1=btreeio.size();
	uint size2=cfltio.size();
	uint size3=totalio.size();
	double avgio=(double)filesize/(double)count;

	uint index1, index2;
	index1=(uint)(0.05*(double)size1);
	index2=(uint)(0.95*(double)size1);
	if (size1>0)
		cout<<" "<<btreeio[index1]<<" "<<btreeio[index2];
	else	
		cout<<" 0 0";
	index1=(uint)(0.05*(double)size2);
	index2=(uint)(0.95*(double)size2);
	if (size2>0)
		cout<<" "<<cfltio[index1]<<" "<<cfltio[index2];
	else	
		cout<<" 0 0";
	index1=(uint)(0.05*(double)size3);
	index2=(uint)(0.95*(double)size3);
	if (size3>0)
		cout<<" "<<totalio[index1]<<" "<<totalio[index2]<<" "<<avgio-(double)totalio[index1]<<" "<<(double)totalio[index2]-avgio<<endl;
	else	
		cout<<" 0 0 0 0"<<endl;
		
  /*vector <u_btree_t *>::iterator it;
  for(it=ubt.begin();it!=ubt.end();++it)
    {
      delete (*it);
    }*/
  
  return 0;
}
