//
// File:    test_ami_btree.cpp
// Author:  Octavian Procopiuc <tavi@cs.duke.edu>
// 
// Test file for class AMI_btree.
//

#include "app_config.h"
#include <portability.h>
#include <versions.h>

/***************************/
#include <sstream>
#include <fstream>
#include <fcntl.h>
#include <unistd.h>
#include <vector>
#include <iostream>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <dirent.h>
#include "mtimer.h"
#include "tools.h"
#include "segment.h"
#include "sdb.h"
/**************************/

VERSION(test_ami_btree_cpp, "$Id: test_ami_btree2.cpp,v 1.7 2009/03/17 06:08:28 lifeifei Exp $");

#include <cpu_timer.h>
#include <ami_btree.h>
#include <ami_stream.h>
#define SIZE_OF_STRUCTURE 2048

// Key type.
//typedef TPIE_OS_OFFSET bkey_t;
typedef double bkey_t;// le

// Element type for the btree.
struct el_t {
  bkey_t key_;
  char data_[SIZE_OF_STRUCTURE-sizeof(bkey_t)];
  
  el_t(bkey_t k): key_(k) 
  {
  }

  el_t() {}
};

struct key_from_el {
  bkey_t operator()(const el_t& v) const { return v.key_; }
};
// Temporary distinction btw UN*X and WIN, since there are some
// problems with the MMAP collection implementation.
#ifdef _WIN32
typedef AMI_btree< bkey_t,el_t,less<bkey_t>,key_from_el,BTE_COLLECTION_UFS > u_btree_t;
#else
typedef AMI_btree< bkey_t,el_t,less<bkey_t>,key_from_el > u_btree_t;
#endif

typedef AMI_STREAM< el_t > stream_t;

// Template instantiations (to get meaningful output from gprof)
//template class AMI_btree_node<bkey_t,el_t,less<bkey_t>,key_from_el>;
//template class AMI_btree_leaf<bkey_t,el_t,less<bkey_t>,key_from_el>;
//template class AMI_btree< bkey_t,el_t,less<bkey_t>,key_from_el,BTE_COLLECTION_UFS >;
//template class AMI_STREAM< el_t >;


// This is 2**31-1, the max value returned by random().
#define MAX_RANDOM ((double)0x7fffffff)
// This is the max value that we want.
#define MAX_VALUE 1000000000
// The number of deletions to perform.
#define DELETE_COUNT 500

// Global variables.
TPIE_OS_OFFSET bulk_load_count;
TPIE_OS_OFFSET insert_count = 5000;
TPIE_OS_OFFSET range_search_lo = 0;
TPIE_OS_OFFSET range_search_hi = 10000000;

bool rankA(segment s1, segment s2)
{
    if(s1.objid<s2.objid)
      return true;
    else if(s1.objid==s2.objid && s1.segid<s2.segid)
      return true;
    else
      return false;

}

class report
{
public:
  segment s;
  double rank;
  report()
  {
    rank=0;
  }
  void assign(segment ss, double rr)
  {
    s=ss;
    rank=rr;
  }
  
  bool operator < (const report & b) const
  {
	return (!Tools::double_le(b.rank,rank,6));
  }
 
  ~report()
  {}
};

int get_tk(double q, vector <segment> &idlst, int topk,int opflag)
{
  int count=0;
  uint i=0;
  double x,y;
  vector <report> rp;
  report myrp;
  sort(idlst.begin(), idlst.end(),rankA);
  uint lastoid=idlst[0].objid, lastsegid=idlst[0].segid;
  segment s;
  for(uint i=0;i<idlst.size();i++)
    {
      s=idlst[i];
      y=0;
	if (s.objid==lastoid && s.segid==lastsegid)
		continue;
      if(s.start.x<=q&&s.end.x>=q && Tools::lineSegmentIntersection(s.start.x,s.start.y,
							     s.end.x,s.end.y,
							     q,100000000,
							     q,0.5,
							     &x,&y)==true)
	{
	  myrp.assign(s,y);
	  rp.push_back(myrp);
	  count++; 
	}
	lastoid=s.objid;
	lastsegid=s.segid;
    }
  if(count)
    {
      sort(rp.begin(),rp.end());
	//report if necessary
	if (opflag==1)
	{
	      uint i=0;
	      cerr.precision(12);
	      for(i=0;i<rp.size();i++)
		{
	  	  	  cerr<<rp[i].s.objid*pow(10.0,4)+rp[i].s.segid<<" @"<<rp[i].s<<" @"<<rp[i].rank<<endl;
		  if(i==topk-1)
		    break;
		}
	}
    }
  rp.clear();
  return count;
}


int get_filesize(const char * filename)
{
  char n[100]="";
  int size=0;
  struct stat buf;
  sprintf(n,"%s",filename);
  if (stat(n, &buf) < 0) {
    fprintf(stderr, "Cannot stat %s\n", n);                                                                               
    exit(1);
  }
  size=buf.st_size;
  return size;
}


int get_treesize(const char * filename)
{
  char n[100]="";
  int size=0;
  struct dirent *w;
  struct stat buf;
  sprintf(n,"%s/btree.ent.l.blk",filename);
  if (stat(n, &buf) < 0) {
    fprintf(stderr, "Cannot stat %s\n", n);                                                                               
    exit(1);
  }
  size+=buf.st_size;
  sprintf(n,"%s/btree.ent.n.blk",filename);
  if (stat(n, &buf) < 0) {
    fprintf(stderr, "Cannot stat %s\n", n);                                                                               
    exit(1);
  }
  size+=buf.st_size;
  return size;
}

int get_cflsize(const char * filename)
{
  char dir[100];
  sprintf(dir,"%s/C",filename);
  DIR *d;
  long size=0;
  struct dirent *w;
  struct stat buf;
  char n[100]="";
  if (stat(dir, &buf) < 0) {
    return 0;
  }
  else
    {
      if(!S_ISDIR(buf.st_mode))
	{
	  if(S_ISLNK(buf.st_mode))
	    {
	      printf("%-ld\t%-s\n",0,dir);
	      exit(0);
	    }
	  else
	    {
	      printf("%-ld\t%-s\n",(long)(buf.st_blocks/2),dir);
	      exit(0);
	    }
	}//if
    }
  if ((d = opendir(dir)) == NULL) {
    printf("This is not supposed to happen.\n");
    exit(0);
  }
  
  while ((w = readdir(d)) != NULL) {
    sprintf(n, "%s/%s", dir, w->d_name);
      if ((strcmp(w->d_name, ".") == 0) ||
	  (strcmp(w->d_name, "..")==0)) continue;
      if (stat(n, &buf) < 0) {
	continue;
      }
      if ((strstr(w->d_name, "."))!=NULL) 
	  continue;
      size+=(long)buf.st_size;
      if (S_ISDIR(buf.st_mode)) {
	size+=get_cflsize(n);
      }
  }
    //stat(dir, &buf);
      //printf("%-ld\t%-s\n",size+(long)(buf.st_blocks/2),dir);
  closedir(d);
  return size;
}

void covertData(el_t& ss, uint& csize, uint& cdistinct, uint& ctotalfile, vector<uint>& clevel, vector<uint>& cindex)	
{
	char* ptr=ss.data_;	
	memcpy((char*)&csize,ptr, sizeof(uint));
	ptr+=sizeof(uint);

	memcpy((char*)&cdistinct,ptr, sizeof(uint));
	ptr+=sizeof(uint);

	memcpy((char*)&ctotalfile, ptr, sizeof(uint));
	ptr+=sizeof(uint);
	uint level=0, index=0;
	//cout<<"csize "<<csize<<" cdistinct "<<cdistinct<<" "<<ctotalfile<<" "<<ctotalfile;
	for (uint i=0;i<ctotalfile;i++)
	{
		memcpy((char*)&level,ptr, sizeof(uint));
		ptr+=sizeof(uint);

		memcpy((char*)&index, ptr, sizeof(uint));
		ptr+=sizeof(uint);

		//cout<<"level "<<level<<"index "<<index<<" ";
		clevel.push_back(level);
		cindex.push_back(index);
	}
	//cout<<endl;
}

int main(int argc, char *argv[]) {

  AMI_err ae;
  TPIE_OS_OFFSET i;
  TPIE_OS_OFFSET j;
  cpu_timer wt;
  int treesize=0,cflsize=0;
  char *base_file = NULL;
  char * base_file_stream = NULL;
  uint T,N=0,k,slvl;
  bool trick_on=false;
  if(argc!=8)
    {
      cerr<<"Usage:test_ami_btree <query input filename> <trick #> <Page Size> <force lvl:S0 is 1> <datafile> <k> <opflag>"<<endl<<flush;
      exit(0);
    }

  //N=atoi(argv[3]);
  T=atoi(argv[2]);
  /*********build conflict list**************************/

  char datafile_folder[100]="";
  strcpy(datafile_folder,argv[5]);
  strcat(datafile_folder,"Index/");
  char buf[100]="";
  stringstream conv,conv1;
  vector<u_btree_t *> ubt;
  char folder[100]="";
  bool flag=false;
  int pagesize=atoi(argv[3]);
  int Bsize=(int)((double)pagesize*1024.0/(double)40); 
  int opflag=atoi(argv[7]);
  //cerr<<Bsize<<endl;
  int force=atoi(argv[4]);
  fstream ifs;
  sprintf(buf,"%sconfig",datafile_folder);
  ifs.open(buf,ios::in);
  if(!ifs.is_open())
    {
      cerr<<"cannot open config file"<<endl;
      exit(1);
    }
  else{
    ifs>>slvl>>N;
    //cerr<<slvl<<" "<<N<<endl;
  }
  ifs.close();
 
  int top_k=atoi(argv[6]);
  long int *hit= new long int[slvl+1];
  memset(hit,0,sizeof(long int)*(slvl+1));
  //cerr<<slvl<<" B+ trees..."<<endl;
  tpie_log_init(TPIE_LOG_APP_DEBUG); 
  //MM_manager.set_memory_limit(64*1024*1024*1024*1024);
  MM_manager.ignore_memory_limit();
  for(k=0;k<slvl;k++)
    {
      conv.str("");
      conv<<datafile_folder;
      conv<<"S"<<k;      
      strcpy(folder,conv.str().data());
      //cerr<<folder<<endl;
      
      // Log debugging info from the application, but not from the library. 
      

      conv.str("");
      conv<<folder<<"/btree.ent"; 
      strcpy(buf,conv.str().data());
      //cerr<<buf<<endl;
      base_file=const_cast <char*> (buf);      
      
      //////  Testing Bulk loading. ///////
      
      u_btree_t *u_btree;
      AMI_btree_params params;
      params.node_block_factor = pagesize/4;
      params.leaf_block_factor = pagesize/4;
      params.leaf_cache_size = 32;
      params.node_cache_size = 64;

      u_btree = new u_btree_t(base_file, AMI_WRITE_COLLECTION, params);

      if (!u_btree->is_valid()) {
	cerr << argv[0] << ": Error initializing btree. Aborting.\n";
	delete u_btree;
	exit(1);
      }
      ubt.push_back(u_btree);
    }

  ifstream fs;
  bkey_t key;				 
  unsigned long filesize=0;
  int io_from_tree=0;
  fs.open(argv[1]);
  int count=0, total=0;
  MTimer t;
  double jdg=((double)top_k/Bsize)>1?((double)top_k/Bsize):1;
  int m=(int)ceil(log(jdg)/log(2));
   uint csize=0, ctotalfile=0, cdistinct=0;
   vector<uint> clevel; 
   vector<uint> cindex;
	uint scanio=(uint)(ceil((double)N/(double)Bsize));
  t.go();
  fs>>total;
  while(count<total)
    {
      int num=0;
      fs>>key;
      //cout<<"Key:"<<key<<endl<<flush;
      segment t;
      vector <segment> vt;
      if (m>force)
	m=force;
      while(num<top_k){
	u_btree_t *u_btree=NULL;
	csize=0;
	ctotalfile=0;
	cdistinct=0;
	clevel.clear();
	cindex.clear();
	if (m>=ubt.size())
	  {
	    filesize+=scanio;
	    hit[slvl]++;
	    break;
	  }
	u_btree=ubt[m];
	el_t ss;
	if(u_btree->pred(key,ss))				 
	  {
	    //cout<<"level "<<m<<" key "<<key<<" ss "<<ss.key_<<" data  "<<endl;
	    covertData(ss,csize, cdistinct, ctotalfile, clevel, cindex);	
	    //cout<<"csize "<<csize<< "cdistinct "<<cdistinct<<" ctotalfile "<<ctotalfile<<" level&index size "<<clevel.size()<<endl;
	    if(csize<top_k || ( csize>T*Bsize*pow(2.0,m) && T>0 ) )
	      {
		num=-1;		  
		filesize+=u_btree->height();
		io_from_tree+=u_btree->height();
		if (m==force)
		  {
		    filesize+=scanio;
		    hit[m]++;
		    break;
		  }
	      }
	    else
	      {
		vt.clear();
		//cout<<"Before read:"<<endl<<flush;
	        uint cios=0;
		SDB::readNConflictList(vt, clevel, cindex, datafile_folder, cios, (uint)Bsize);		
		//cout<<"after read,size:"<<vt.size()<<endl<<flush;
		num=vt.size();
		io_from_tree+=u_btree->height();
		filesize=filesize+cios+u_btree->height();
		int a=0;
		a=get_tk(key,vt,top_k,opflag);
		if(a<top_k)
		  {
		    num=-1;
		    if (m==force)
		      {
			filesize+=scanio;
			hit[m]++;
			break;
		      }
		  }
		else
		  {
		       hit[m]++;
		       break;
		  }
	      }
	  }
	    else
	      break;
	m++;
      }
      count++;
    }          
  fs.close();
  t.stop();
  t.update();
  /*avg time on query, avg cost on tree traverse, avg cost on reading CFL, SUM of the previous two, hit[...], trick, force*/
  cout<<t.elapsed/count<<" "<<(double)io_from_tree/count<<" "<<(double)filesize/count-(double)io_from_tree/count
      <<" "<<(double)filesize/count<<" ";
  int iter=0;
  for(;iter<=ubt.size();iter++)
    cout<<(double)hit[iter]/count<<" ";
  cout<<T<<" "<<pagesize<<" "<<force<<" "<<top_k<<endl;
  
  
  vector <u_btree_t *>::iterator it;
  for(it=ubt.begin();it!=ubt.end();++it)
    {
      delete (*it);
    }
  
  return 0;
}
